Example: vertexArrays.c

/*  vertexArrays.c  - vertex array example. uses the same shaded
 *                    polygon seen in ~opengl/rendering/shadeModel.c
 *
 *     SPACE key       - toggle between flat/smooth shading
 *     ESCAPE key      - exit program
 */

#include <GL/glut.h>   /* includes gl.h, glu.h */

#include <stdio.h>

/*  Function Prototypes  */

GLvoid  initgfx( GLvoid );
GLvoid keyboard( GLubyte, GLint, GLint );
GLvoid  drawScene( GLvoid );

void printHelp( char * );

/* Global Definitions */

#define KEY_ESC        27      /* ascii value for the escape key */

static GLint flatShading = GL_FALSE;

void
main( int argc, char *argv[] )
{
    GLsizei width, height;

    glutInit( &argc, argv );

    width = glutGet( GLUT_SCREEN_WIDTH ); 
    height = glutGet( GLUT_SCREEN_HEIGHT );
    glutInitWindowPosition( width / 4, height / 4 );
    glutInitWindowSize( width / 2, height / 2 );
    glutInitDisplayMode( GLUT_RGBA );
    glutCreateWindow( argv[0] );

    initgfx();

    glutKeyboardFunc( keyboard );
    glutDisplayFunc( drawScene ); 

    printHelp( argv[0] );

    glutMainLoop();
}

void
printHelp( char *progname )
{
    fprintf(stdout, "\n%s - demonstrates vertex arrays\n\n"
            "SPACE Key    - toggle between flat/smooth shading\n"
            "Escape Key   - exit the program\n\n",
            progname);
}

GLvoid
initgfx( GLvoid )
{
    glClearColor( 0.0, 0.0, 1.0, 1.0 );
    glLineWidth( 4.0 );
 
    glOrtho( 0.0, 2.1, 0.0, 2.1, -2.0, 2.0 );

#ifndef GL_VERSION_1_1
#ifndef GL_EXT_vertex_array
    printf("vertex arrays are not supported on this system\n");
#endif /* GL_EXT_vertex_array */
#endif /* !GL_VERSION_1_1 */
}

GLvoid 
keyboard( GLubyte key, GLint x, GLint y )
{
    switch (key) {
    case ' ':
         flatShading = !flatShading;
         if (flatShading) 
              glShadeModel(GL_FLAT);
         else
              glShadeModel(GL_SMOOTH);
         glutPostRedisplay();
         break;
    case KEY_ESC:   /* Exit whenever the Escape key is pressed */
         exit(0);
    }
}

GLvoid
drawVertices( GLvoid )
{
    static GLfloat    purpleColor[] = { 1.0, 0.0, 1.0 };
    static GLfloat    redColor[]  = { 1.0, 0.0, 0.0 };
    static GLfloat    blackColor[] = { 0.0, 0.0, 0.0 };
    static GLfloat    darkgreenColor[] = { 0.0, 0.5, 0.0 };
    static GLfloat    whiteColor[] = { 1.0, 1.0, 1.0 };

    /* Gouraud shade a polygon */
    glBegin( GL_POLYGON );
      glColor3fv( purpleColor );
      glVertex2f( 0.5, 0.1 );
      glColor3fv( redColor );
      glVertex2f( 1.0, 0.4 );
      glColor3fv( blackColor );
      glVertex2f( 0.9, 1.0 );
      glColor3fv( darkgreenColor );
      glVertex2f( 0.3, 0.8 );
      glColor3fv( whiteColor );
      glVertex2f( 0.1, 0.5 );
    glEnd();
}

GLvoid
drawArrays( GLvoid )
{
    static GLfloat  colorArray[] = {
         1.0, 0.0, 1.0, /* purple */
         1.0, 0.0, 0.0, /* red */
         0.0, 0.0, 0.0, /* black */
         0.0, 0.5, 0.0, /* dark green */
         1.0, 1.0, 1.0  /* white */
    };

    static GLfloat  vertexArray[] = {
         0.5, 0.1,
         1.0, 0.4,
         0.9, 1.0,
         0.3, 0.8,
         0.1, 0.5
    };

    static GLuint   indices[] = { 2, 3, 4, 0, 1 };
    int i;
   
#ifdef GL_VERSION_1_1 /* OpenGL version 1.1 vertex array API */

    /* set up the array data */
    glColorPointer( 3, GL_FLOAT, 3*sizeof(GLfloat), colorArray );
    glVertexPointer( 2, GL_FLOAT, 2*sizeof(GLfloat), vertexArray );

    /* enable vertex arrays */
    glEnableClientState( GL_COLOR_ARRAY );
    glEnableClientState( GL_VERTEX_ARRAY );

    /* draw a polygon using the arrays sequentially */
    glDrawArrays( GL_POLYGON, 0, 5 );

    /* draw a line loop using the same data */
    glPushMatrix();
         glTranslatef( 1.0, 0.0, 0.0 );
         glDrawArrays ( GL_LINE_LOOP, 0, 5 );

         /* draw a triangle fan using the arrays in a different order */
            glPushMatrix();
            glTranslatef( 0.0, -1.0, 0.0 );
            glDrawElements ( GL_TRIANGLE_FAN, 5, GL_UNSIGNED_INT, indices );
         glPopMatrix();
    glPopMatrix();

    glDisableClientState( GL_COLOR_ARRAY );
    glDisableClientState( GL_VERTEX_ARRAY );

#else /* !GL_VERSION_1_1 */
#ifdef GL_EXT_vertex_array /* OpenGL version 1.0 vertex array extension */

    /* set up the array data */
    glColorPointerEXT( 3, GL_FLOAT, 3*sizeof(GLfloat), 5, colorArray );
    glVertexPointerEXT( 2, GL_FLOAT, 2*sizeof(GLfloat), 5, vertexArray );

    /* enable vertex arrays */
    glEnable( GL_COLOR_ARRAY_EXT );
    glEnable( GL_VERTEX_ARRAY_EXT );

    /* draw a polygon using the arrays sequentially */
    glDrawArraysEXT( GL_POLYGON, 0, 5 );

    /* draw a line loop using the same data */
    glPushMatrix();
        glTranslatef( 1.0, 0.0, 0.0 );
        glDrawArraysEXT( GL_LINE_LOOP, 0, 5 );

        /* draw a triangle fan using the arrays in a different order */
        glPushMatrix();
                glTranslatef( 0.0, -1.0, 0.0 );
                glBegin( GL_TRIANGLE_FAN );
                    for (i = 0; i < 5; i++)
                        glArrayElementEXT( indices[i] );
                glEnd();
        glPopMatrix();
    glPopMatrix();

    glDisable( GL_COLOR_ARRAY_EXT );
    glDisable( GL_VERTEX_ARRAY_EXT );

#endif /* GL_EXT_vertex_array */
#endif /* GL_VERSION_1_1 */
}

GLvoid
drawScene( GLvoid )
{
    glClear( GL_COLOR_BUFFER_BIT );

    /* draw lower left polygon without vertex arrays */
    drawVertices();

    glPushMatrix();
         glTranslatef( 0.0, 1.0, 0.0 );

         /* draw the same arrays three different ways */
         drawArrays();
    glPopMatrix();

    glFlush();
}


