#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <SDL.h>
#include <GL/gl.h>
#include <GL/glu.h>

#include "bitmap.h"
#include "texture.h"

#ifndef M_PI
#define M_PI           3.14159265358979323846  /* pi */
#endif

/* creates pyramid with given count of sides and no bottom */
typedef GLfloat GLfloattriplet[3];
typedef GLfloat GLfloatdoublet[2];
typedef GLubyte GLubytetriplet[3];

/*
int pyramid(int sides, GLfloat radius, GLfloat height,
  GLfloattriplet **vertices, GLubytetriplet **colors, GLfloatdoublet **texc) {
  if (vertices == NULL || colors == NULL) return -1;
  int i;
  GLfloat angle;
  *vertices = (GLfloattriplet *) malloc((2+sides) * sizeof(GLfloattriplet));
  *colors   = (GLubytetriplet *) malloc((2+sides) * sizeof(GLubytetriplet));
  *texc     = (GLfloatdoublet *) malloc((2+sides) * sizeof(GLfloatdoublet));

  (*vertices)[0][0] = 0.0f;
  (*vertices)[0][1] = 0.0f;
  (*vertices)[0][2] = height;

  (*texc)[0][0] = 0.0f;
  (*texc)[0][1] = 0.0f;
  (*texc)[1][0] = 1.0f;
  (*texc)[1][1] = 0.0f;
  
  for (i = 0; i <= sides; i++) {
    (*vertices)[i+1][0] = radius*cos(angle = 2*M_PI * i / sides);
    (*vertices)[i+1][1] = radius*sin(angle);
    (*vertices)[i+1][2] = 0.0f;
    if (i > 0) {
      (*texc)[i+1][0] = 0.0f;
      (*texc)[i+1][1] = 1.0f;
    }
  }
  return 0;
}*/

int prism(int sides, GLfloat radius, GLfloat height,
  GLfloattriplet **vertices, GLubytetriplet **colors, GLfloatdoublet **texc) {
  if (vertices == NULL || colors == NULL || texc == NULL) return -1;
  int i;
  GLfloat angle;
  *vertices = (GLfloattriplet *) malloc(2*(1+sides) * sizeof(GLfloattriplet));
  *colors   = (GLubytetriplet *) malloc(2*(1+sides) * sizeof(GLubytetriplet));
  *texc     = (GLfloatdoublet *) malloc(2*(1+sides) * sizeof(GLfloatdoublet));

  for (i = 0; i <= sides; i++) {
    (*vertices)[2*i][0] = radius*cos(angle = 2*M_PI * i / sides);
    (*vertices)[2*i][1] = radius*sin(angle);
    (*vertices)[2*i][2] = 0.0f;
    (*vertices)[2*i+1][0] = radius*cos(angle = 2*M_PI * i / sides);
    (*vertices)[2*i+1][1] = radius*sin(angle);
    (*vertices)[2*i+1][2] = height;
   
    (*texc)[2*i][0] = 0.0;
    (*texc)[2*i][1] = 0.0;
    (*texc)[2*i+1][0] = 0.0;
    (*texc)[2*i+1][1] = 1.0;
  }
  return 0;
}

struct options {
  SDL_Surface *s;
  int running;

  int arrays;
  int complexity;
  GLfloattriplet *vertices;
  GLubytetriplet *colors;
  GLfloatdoublet *texc;
  BITMAP *texture;
  GLuint textureid;
  unsigned long int frames;
  unsigned long int fftime;
};

int core_handleevent(struct options *options) {
  SDL_Event event;

  while( SDL_PollEvent( &event ) ) {
    switch( event.type ) {
      case SDL_KEYDOWN:
        switch (event.key.keysym.sym) {
          case SDLK_ESCAPE:
            options->running = 0;
            break;
          default: break; /* cuts the compiler's warning crap */
        }
        break;
      case SDL_QUIT:
        options->running = 0;
        break;
      default: break;
    }
  }
  return 0;
}


static void draw_screen(struct options *options) {
  static float angle = 0.0f;
  
  /* Clear the color and depth buffers. */
  glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

  /* We don't want to modify the projection matrix. */
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  /* Move down the z-axis. */
  glTranslatef(0.0, 0.0, -5.0);

  /* Rotate. */
  glRotatef(angle, 1.0, 1.0, 1.0);

  if( (angle += 1.0) > 360.0f ) {
    angle = 0.0f;
  }

  glBindTexture(GL_TEXTURE_2D, options->textureid);
  if (options->arrays) {
    glVertexPointer(3, GL_FLOAT, 0, options->vertices);
    glTexCoordPointer(2, GL_FLOAT, 0, options->texc);
    glEnableClientState(GL_VERTEX_ARRAY);
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
    glDrawArrays(GL_QUAD_STRIP, 0, 2*(1+options->complexity));
  } else {
    glBegin(GL_QUAD_STRIP);
    int i;
    for (i = 0; i < 2 * (1+options->complexity); i++) {
      glTexCoord2fv(options->texc[i]);
      glVertex3fv(options->vertices[i]);
    }
    glEnd();
  }
  SDL_GL_SwapBuffers( );
}

int core(struct options *options) {
  if (options == NULL) return -1;
  unsigned long int f1, f2;
  options->running = 1;
  options->frames = 0;
  options->fftime = f1 = SDL_GetTicks();
  double d;

  while (options->running) {
    core_handleevent(options);
    draw_screen(options);
    f2 = SDL_GetTicks();
    if (f2 - f1 >= 1000) {
      d = (options->frames) * 1000.0 / (f2 - options->fftime);
      printf("\r %.1f frames -- %.1f rectangles/s -- %.1f triangles.", d,
        options->complexity*d, options->complexity*d*2);
      f1 = f2;
    }
    ++(options->frames);
  }
  printf("\n");

  return 0;
}

static void setup_opengl( int width, int height )
{
  float ratio = (float) width / (float) height;

  glEnable(GL_TEXTURE_2D);
  glShadeModel(GL_SMOOTH);
  glEnable(GL_DEPTH_TEST);
  glDepthFunc(GL_LEQUAL);
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
 
  /* Culling. */
//  glCullFace( GL_BACK );
//  glFrontFace( GL_CCW );
//  glEnable( GL_CULL_FACE );

  /* Set the clear color. */
  glClearColor( 0, 0, 0, 0 );

  /* Setup our viewport. */
  glViewport(0, 0, width, height);

  /*
   * Change to the projection matrix and set
   * our viewing volume.
   */
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(60.0, ratio, 1.0, 1024.0);
}

int setup_objects(struct options *options, int complexity)
{
  if (options == NULL) return -1;
  printf("(level %d -> %d-sided prism) ", complexity,
    options->complexity = complexity + 3);
  if (prism(options->complexity, 1, 1,
    &options->vertices, &options->colors, &options->texc) < 0) return -1;
 
  return 0;
}

void getopenglinfo(struct options *options) {
  const char *glVendor = (const char *) glGetString(GL_VENDOR);
  const char *glRenderer = (const char *) glGetString(GL_RENDERER);
  const char *glVersion = (const char *) glGetString(GL_VERSION);
  const char *glExtensions = (const char *) glGetString(GL_EXTENSIONS);
  printf(
    "OpenGL vendor     : \"%s\"\n"
    "OpenGL renderer   : \"%s\"\n"
    "OpenGL version    : \"%s\"\n"
    "OpenGL extensions : \"%s\"\n", glVendor, glRenderer, glVersion,
    glExtensions);
  options->arrays = strstr(glExtensions, "GL_EXT_multi_draw_arrays") != NULL
    && strstr(glExtensions, "GL_EXT_vertex_array") != NULL;
}

int loadgraphics(struct options *options) { 
  if (options == NULL) return -1;
  options->texture = createbitmap2(256, 256, 32, 4);
  unsigned long int x, y;
  for (y = 0; y < 256; y++)
    for (x = 0; x < 256; x++)
      b_putpixel(options->texture, x, y,
        (x >> 6) & (y >> 6) & 1 ? 0xffffff :
        (x >> 6) & 1 ? 0x00ff00 : (y >> 6) & 1 ? 0xff0000 : 0x0000ff);
//      b_putpixel(options->texture, x, y, ((x % 64) << 2) |
//        ((y % 64) << 10));
  return 0;
}

void setuptextures(struct options *options) {
  if (options == NULL) return;
  options->textureid = createtexturefrombitmap(options->texture);
}

static struct options options;

void shutdown_function() {
  freebitmap(options.texture);
  printf("Closing down SDL...\n");
  SDL_Quit();
}

int main(int argc, char **argv) {
  if (argc < 2) {
    printf("syntax: %s complexity_integer\n", argv[0]);
    return 1;
  }

  int complexity;
  char *p;
  complexity = strtol(argv[1], &p, 0);
  if (p == argv[1] || *p != '\0') {
    printf("'%s' is not a valid integer\n", argv[1]);
    return 2;
  }
  if (complexity < 0) {
    printf("Complexity is negative -> * (-1)\n");
    complexity *= -1;
  }
  
  printf("Setting up SDL: ");
  if (SDL_Init(SDL_INIT_VIDEO) == -1) {
    printf("%s\n", SDL_GetError());
    return 3;
  }
  printf("OK.\n");
  atexit(shutdown_function);

  const SDL_VideoInfo *info;

  if ((info = SDL_GetVideoInfo()) == NULL) {
    printf("I can't get information about graphic subsystem.\n");
    return 4;
  }

  int sdlflags = 0, width, height, bpp;
  
  /* We want at least double-buffered 16-bit Z-buffer */
  SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 5);
  SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 5);
  SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, 5);
  SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 16);
  SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);

  width = 640;
  height = 480;
  bpp = info->vfmt->BitsPerPixel;
  sdlflags |= SDL_OPENGL | SDL_FULLSCREEN;
  
  printf("Loading textures: ");
  if (loadgraphics(&options) < 0) {
    printf("failed.\n");
    return 5;
  } else printf("OK.\n");
  
  printf("Preparing objects: ");
  if (setup_objects(&options, complexity) < 0) {
    printf("failed.\n");
    return 6;
  } else printf("OK.\n");
  printf("Setting video mode (%dx%dx%d): ", width, height, bpp);
  if ((options.s = SDL_SetVideoMode(width, height, bpp, sdlflags)) != NULL) {
    printf("OK.\n");
    setup_opengl(width, height);
    printf("OpenGL information goes here... \n");
    getopenglinfo(&options);
    printf("... end of OpenGL information.\n");
    printf("We %s use vertex arrays!\n", options.arrays ? "can" : "can't");
    printf("Configuring textures...\n");
    setuptextures(&options);
    core(&options);
    SDL_FreeSurface(options.s);
  } else {
    printf("%s\n", SDL_GetError());
    return 7;
  }
 
  return 0;
}
